
/*
 *  H45 loader
 *  Thu Apr 19 11:01:54 EEST 2007
 *
 *  Thanks to OH3NWQ
 *
 *  Connector pins, looking from outside the radio
 *  (can you fit inside it?)
 *
 *                       o   o  RXD ---o<--- PC TXD
 *                         o
 *                  GND  o   o
 *                         o
 *  PC RXD ---o<--- TXD  o   o
 */

#define SIZE_BOOT  2656 /* boot.bin really not fixed size ? */
#define SIZE_APP 0x8000

static char
	*dev_path  = "/dev/ttyd1",
	*boot_path,              /* use builtin boot.bin */
	*firm_path = "hd45.bin";

static int
	zap_settings;            /* last 0x1000 bytes sent also */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

typedef unsigned char byte;

static int  dev;
static byte boot_bin[];

static void
oops(char *with)
{
	perror(with);
	exit(1);
}

static void
put(byte b)
{
	if (write(dev, &b, 1) != 1)
		oops("write");
}

static byte
get(void)
{
	byte b;

	fflush(stdout);

	if (read(dev, &b, 1) != 1)
		oops("read");

	return (b);
}

static void
show(byte b)
{
	if (b > ' ' && b < 0x7F && b != '\\')
		printf("%c", b);
	else
		printf("\\%02X", b);

	fflush(stdout);
}

static void
say(byte b)
{
	printf(" ");
	show(b);
	printf("! ");

	put(b);
}

static void
expect(byte it)
{
	printf(" ");
	show(it);
	printf("? ");

	for (;;) {
		byte b = get();

		if (b == it)
			break;

		show(b);
	}
}

static void
load(FILE *from, int binary, int size)
{
	int dot = 0, c, chunk = 0;

	do {
		if (from)
			c = getc(from);
		else
			c = boot_bin[dot];

		if (c == EOF) {
			printf(" short file\n");
			break;
		}
		if (binary) {
			put(c);

			if (++chunk == 32) {
				chunk = 0;

				c = get();
				if (c != 'B') {
					show(c);
					printf("\nFailed\n");
					exit(1);
				}
				show('.');
			}
		} else {
			byte xx[4];

			sprintf(xx, "%02X", c);
			put(xx[1]);              /* reverse order, funny. */
			put(xx[0]);

			if (++chunk == 128) {
				chunk = 0;

				show('.');
			}
		}
	} while (++dot != size);
}

main(int argc, char **argv)
{
	extern int optind, opterr;
	extern char *optarg;

	int c;

	FILE
		*boot,
		*firm;

	while ((c = getopt(argc, argv, "zb:f:d:")) != -1)
		switch (c) {
		default:
	baduse:
			fprintf(stderr, "\
Usage: h45loader -b bootfile -f firmfile -d ttydev -z(ap settings)\n");
			exit(2);

		case 'b':
			boot_path = optarg;
			break;

		case 'f':
			firm_path = optarg;
			break;

		case 'd':
			dev_path = optarg;
			break;

		case 'z':
			zap_settings = 1;
			break;
		}

	if (optind != argc)
		goto baduse;

	if (boot_path) {
		boot = fopen(boot_path, "r");
		if (!boot)
			oops(boot_path);
	} else {
		boot_path = "(builtin boot.bin)";
		boot = NULL;
	}

	firm = fopen(firm_path, "r");
	if (!firm)
		oops(firm_path);

	{
		struct stat st;

		if (boot) {
			if (fstat(fileno(boot), &st) == -1)
				oops(boot_path);

			if (st.st_size != SIZE_BOOT)
				printf("%s not %d bytes\n", boot_path, SIZE_BOOT);
		}

		if (fstat(fileno(firm), &st) == -1)
			oops(firm_path);

		if (st.st_size != SIZE_APP)
			printf("%s not %d bytes\n", firm_path, SIZE_APP);
	}
	{
		struct termios tt, ttsave;

		dev = open(dev_path, O_RDWR | O_NONBLOCK);
		if (dev == -1)
			oops(dev_path);

		if (tcgetattr(dev, &tt) == -1)
			oops(dev_path);

		ttsave = tt;

		cfmakeraw(&tt);
		cfsetispeed(&tt, B9600);
		cfsetospeed(&tt, B9600);

		tt.c_cflag |= CLOCAL;

		tt.c_iflag |= PARMRK;

		if (tcsetattr(dev, TCSAFLUSH, &tt) == -1)
			oops(dev_path);

		fcntl(dev, F_SETFL, ~O_NONBLOCK & fcntl(dev, F_GETFL, 0));
	}

	printf("Waiting\n");

	expect('Q');
	put('q');        /* not "saying", but replying to Q */

	say('M');        /* 'Monitor load' ? */
	expect('m');
	expect('m');

	printf("\nSending %s\n", boot_path);

	sleep(1);
	load(boot, 0, SIZE_BOOT);

	sleep(1);
	say(0x03);           /* ETX for boot.bin */
	expect(0x03);
	expect(0x03);

	sleep(1);
	say('G');            /* 'Go' ? */
	expect('g');

	sleep(1);
	say('r');            /* ready ? reset ? */
	expect('R');

	sleep(1);
	say('A');            /* 'Application load' ? */
	expect('a');

	printf("\nSending %s\n", firm_path);

	sleep(1);
	load(firm, 1, zap_settings ? SIZE_APP : 0x7000);

	sleep(1);
	if (zap_settings) {
		say(0x03);       /* boot.bin responds to ETX only after SIZE_APP */
		expect(0x03);    /* XXX remove this etxetx altogether ? */
	}
	printf("\nThumbs up!\n");

#if 1
	for (;;) {
		byte b = get();

		show(b);
	}
#endif
	exit(0);
}

/* nothing after this large table */

static
byte
boot_bin[] = {
	142, 55,255,206, 48, 11,140, 50,118, 39,  5,111,  0,  8, 32,246,
	206, 48,  0,223, 64,206, 42, 85,230,  0,  8, 60,222, 64,140, 48,
	 11, 39,  8,231,  0,  8,223, 64, 56, 32,237, 56,206,  0, 64,223,
	 64,206, 42, 81,230,  0,  8, 60,222, 64,140,  0, 64, 39,  8,231,
	  0,  8,223, 64, 56, 32,237, 56,189, 41, 44,126, 32,  0, 60, 48,
	237,  0, 32, 21, 48,238,  0,230,  0,189, 35,236, 48,108,  1, 38,
	  2,108,  0,204,  0,  5,189, 34,159, 48,238,  0,109,  0, 38,228,
	 56, 57, 48, 24,195,255,244, 24, 53,237,  0,206,  0,  5, 60, 48,
	 24,195,  0,  4, 55, 54, 48,236,  4,189, 33, 97, 56, 56,131,  0,
	  0, 39,  7, 48, 24,195,  0,  2,141,180, 48, 24,195,  0, 12, 24,
	 53, 57, 60, 60, 60, 52, 48,237,  0,230, 10,134,  0, 55, 54, 48,
	 24,195,  0,  4, 55, 54, 48,230,  5,134,  0,189, 33, 97, 56, 56,
	131,  0,  0, 39,  7, 48, 24,195,  0,  2,141,130, 56, 56, 56, 49,
	 57, 60, 60, 60, 52,221, 66,127,  0, 64, 48,111,  5,111,  6, 32,
	 11,124,  0, 64,124,  0, 67, 38,  3,124,  0, 66,222, 66,109,  0,
	 38,239,220, 66,131,  0,  1,221, 66,204,  0,  1, 48,237,  3,214,
	 64,231,  0,125,  0, 64, 39, 82,111,  1,111,  2, 32, 63,222, 66,
	230,  0,193, 48, 45, 41,193, 57, 46, 37, 48,236,  3,221, 64,222,
	 66,230,  0, 23, 72,134,  0,130,  0,195,255,208,189, 41, 81, 48,
	227,  5,237,  5,236,  3,  5,221, 64,  5,  5,211, 64,237,  3,220,
	 66,131,  0,  1,221, 66, 48,108,  2, 38,  2,108,  1, 48,230,  0,
	134,  0,221, 64,236,  1,147, 64, 37,180,236,  5, 56, 56, 56, 49,
	 57, 48, 24,195,255,246, 24, 53,237,  4, 55, 54, 56,140,255,252,
	 35, 11,204,  0,  0, 48, 24,195,  0, 10, 24, 53, 57, 48,236, 12,
	237,  6,204,  0,  1,221, 66,204,  0,  1,237,  8,236,  4,237,  2,
	 32, 29,236,  8,  5,221, 64,  5,  5,211, 64,237,  8,221, 64,236,
	  4,189, 41,192, 48,237,  2,124,  0, 67, 38,  3,124,  0, 66, 48,
	236,  2,131,  0,  9, 34,219,236, 14,147, 66, 47, 44,236, 14,147,
	 66,237,  0,127,  0, 66,127,  0, 67, 32, 21,198, 48,238,  6,231,
	  0, 48,108,  7, 38,  2,108,  6,124,  0, 67, 38,  3,124,  0, 66,
	 48,220, 66,163,  0, 45,228, 32, 59,127,  0, 66,127,  0, 67, 32,
	 51,221, 64,236,  4,189, 41,192,203, 48, 48,238,  6,231,  0, 48,
	108,  7, 38,  2,108,  6,236,  8,221, 64,198,  4, 58,189, 42, 67,
	204,  0, 10,221, 64, 48,198,  8, 58,189, 42, 53,124,  0, 67, 38,
	  3,124,  0, 66, 48,236,  8, 38,200,238,  6,111,  0,220, 66,126,
	 33,117,198,  1,247, 48, 11, 57,127, 50, 13,127, 49, 16,127, 49,
	 18,127, 49, 17,127, 50, 15,127, 50, 14,198,  5,215, 16,127,  0,
	 30,204,  0,  6,141, 73,127, 48, 13,114,  8, 17,114,  2, 17,114,
	 16, 17, 57,214, 20,247, 48, 17,198, 96,250, 48, 17,247, 48, 17,
	196,224,215, 20,198,255,215, 22,198,128,215, 32,198,244,215,  1,
	204,  0,  5,141, 26,198,  4,215,  3,127,  0, 23,127, 48, 16,198,
	  4,247, 48, 14,189, 36,174,141,159, 57,198,  7,141, 80, 57, 60,
	 60, 48,237,  0,204,  0,  0, 32, 32,127,  0, 66,127,  0, 67, 32,
	 12,220, 66,221, 64,124,  0, 67, 38,  3,124,  0, 66, 48,220, 66,
	163,  0, 37,237,236,  2,195,  0,  1,237,  2, 55, 54, 56,140,  0,
	 15, 45,214, 56, 56, 57,215, 67,150, 67,198,  1, 77, 39,  4, 88,
	 74, 38,252,215, 66,250, 48, 16,247, 48, 16,215, 23, 57,215, 67,
	150, 67,198,  1, 77, 39,  4, 88, 74, 38,252, 83,215, 66,244, 48,
	 16,247, 48, 16,215, 23, 57,215, 67,150, 67,198,  1, 77, 39,  4,
	 88, 74, 38,252,215, 66,250, 48, 15,247, 48, 15,215, 21, 57,215,
	 67,150, 67,198,  1, 77, 39,  4, 88, 74, 38,252, 83,215, 66,244,
	 48, 15,247, 48, 15,215, 21, 57,221, 64,214, 65,193,  0, 39, 18,
	193,  1, 38,  8,198,128,250, 48, 14,247, 48, 14,246, 48, 14,215,
	  3, 57,198,127,244, 48, 14, 32,240,221, 64, 55, 54, 56,140,  0,
	  1, 39,  6,140,  0,  2, 39,  7, 57, 48,236,  2,221, 11, 57, 48,
	236,  2,221, 25, 57,150, 17,125, 48, 13, 39, 49,133, 32, 39, 48,
	113,251, 17,125, 50, 13, 39, 32,206, 49, 19,246, 50, 14, 58,230,
	  0,215, 19,124, 50, 14,246, 50, 14,193,250, 38,  3,127, 50, 14,
	122, 50, 13,114,  4, 17, 32,  8,127, 48, 13, 32,  3,113,251, 17,
	133,128, 39, 46,113,239, 17,214, 18,247, 48, 12,246, 49, 16,193,
	254, 36, 28,124, 49, 16,206, 48, 18,246, 49, 18, 58,246, 48, 12,
	231,  0,124, 49, 18,246, 49, 18,193,254, 38,  3,127, 49, 18,114,
	 16, 17,133, 64, 39,  5,214, 18,247, 48, 12, 59,215, 66,113,251,
	 17,246, 50, 13,193,250, 36, 39,206, 49, 19,246, 50, 15, 58,214,
	 66,231,  0,124, 50, 13,124, 50, 15,246, 50, 15,193,250, 38,  3,
	127, 50, 15,198,  1,247, 48, 13,114,  4, 17,204,  0,  1, 57,114,
	  4, 17,204,  0,  0, 57,221, 64,113,239, 17,125, 49, 16, 39, 35,
	206, 48, 18,246, 49, 17, 58,230,  0,222, 64,231,  0,122, 49, 16,
	114, 16, 17,124, 49, 17,246, 49, 17,193,254, 38,  3,127, 49, 17,
	198,  1, 57,114, 16, 17,198,  0, 57, 59,214, 20,247, 48, 17,198,
	  1,250, 48, 17,247, 48, 17,215, 20, 57,214, 20,247, 48, 17,198,
	254,244, 48, 17,247, 48, 17,215, 20, 57, 60, 48,237,  0,195,  0,
	 22, 68, 86,221, 64,  5,211, 64,195, 63,208,221, 66,236,  4,222,
	 66,237,  1,198,126,231,  0, 56, 57,198,  6,189, 34,238,198,  5,
	189, 34,214,198,  3,189, 34,238,198,  4,189, 34,238, 57,206, 36,
	 89, 60,204,255,242,141,195, 56,206, 35,117, 60,204,255,240,141,
	185, 56,206, 36, 89, 60,204,255,244,141,175, 56,206, 36, 89, 60,
	204,255,246,141,165, 56,206, 36, 89, 60,204,255,248,141,155, 56,
	206, 36, 89, 60,204,255,250,141,145, 56,206, 36, 89, 60,204,255,
	252,141,135, 56,206, 36, 89, 60,204,255,238,189, 36,122, 56,206,
	 36, 89, 60,204,255,234,189, 36,122, 56,206, 36, 89, 60,204,255,
	236,189, 36,122, 56, 57, 60,221, 64,127,  0, 67,127,  0, 66,126,
	 37,170,222, 64, 58,230,  0,193, 65, 45, 34,222, 64,214, 66, 58,
	230,  0,193, 70, 46, 23,222, 64,214, 66, 58,230,  0,203,201, 55,
	 48,198,  1, 58,214, 66, 58, 51,231,  0,124,  0, 67,222, 64,214,
	 66, 58,230,  0,193, 97, 45, 34,222, 64,214, 66, 58,230,  0,193,
	102, 46, 23,222, 64,214, 66, 58,230,  0,203,169, 55, 48,198,  1,
	 58,214, 66, 58, 51,231,  0,124,  0, 67,222, 64,214, 66, 58,230,
	  0,193, 48, 45, 34,222, 64,214, 66, 58,230,  0,193, 57, 46, 23,
	222, 64,214, 66, 58,230,  0,203,208, 55, 48,198,  1, 58,214, 66,
	 58, 51,231,  0,124,  0, 67,124,  0, 66,214, 66,193,  2, 36,  3,
	126, 37, 34,214, 67,193,  2, 38, 17, 48,230,  1, 88, 88, 88, 88,
	234,  0,238,  4,231,  0,198,  1, 56, 57,198,  0, 32,250, 60,221,
	 64,214, 65, 84, 84, 84, 84, 48,231,  1,214, 65,196, 15,231,  0,
	127,  0, 66, 32, 76, 48,198,  0, 58,214, 66, 58,230,  0,193,  9,
	 34, 17, 48,198,  0, 58,214, 66, 58,166,  0,139, 48, 48,238,  4,
	 58,167,  0, 48,198,  0, 58,214, 66, 58,230,  0,193, 10, 37, 30,
	 48,198,  0, 58,214, 66, 58,230,  0,193, 15, 34, 17, 48,198,  0,
	 58,214, 66, 58,166,  0,139, 55, 48,238,  4, 58,167,  0,124,  0,
	 66,214, 66,193,  2, 37,174, 56, 57, 60, 60,221, 64, 48,237,  0,
	204,  0,  0, 32, 26, 48,238,  0,230,  0,189, 35,236,204,  0, 10,
	189, 34,159, 48,108,  1, 38,  2,108,  0,236,  2,195,  0,  1,237,
	  2,238,  6,223, 64, 55, 54, 56,156, 64, 37,217,204,  0,  0, 32,
	 11,198,255,189, 35,236, 48,236,  2,195,  0,  1, 48,237,  2, 55,
	 54, 56,140,  0,  6, 37,234,198,  3,189, 35,236, 56, 56, 57, 60,
	 60, 60, 48,237,  4,204,  0,  1,189, 35, 56,198,  1, 48,231,  0,
	127, 50, 56,127, 50, 57,127, 50, 52,127, 50, 53,127, 50, 51,236,
	  4,253, 50, 17,111,  2,111,  1, 48,109,  0, 38, 15,204,  0,  0,
	189, 35, 56,198,  3,189, 35,236, 56, 56, 56, 57, 24,195,  0,  3,
	189, 36, 38, 93, 39,226, 48,230,  3,193,  3, 38, 78,109,  1, 39,
	 74,111,  0,125, 50, 51, 39,208,127, 50, 16,113,239, 17,113,251,
	 17,246, 50, 51,193, 32, 35, 29,198, 32,247, 50, 51, 32, 22,206,
	 50, 19, 58,230,  0,254, 50, 17,231,  0,124, 50, 16,124, 50, 18,
	 38,  3,124, 50, 17,246, 50, 16,241, 50, 51, 37,226,204,  0, 50,
	189, 34,159,114, 16, 17,114,  4, 17, 32,141, 48,230,  3,193,255,
	 38,  4,108,  2, 32,  4,111,  2,111,  1,230,  2,193,  6, 38,  4,
	198,  1,231,  1,166,  3,206, 50, 19,246, 50, 51, 58,167,  0,124,
	 50, 51,124, 50, 57, 38,  3,124, 50, 56,246, 50, 51,193, 32, 39,
	  3,126, 38,184,127, 50, 51,113,239, 17,113,251, 17, 32, 22,206,
	 50, 19, 58,230,  0,254, 50, 17,231,  0,124, 50, 51,124, 50, 18,
	 38,  3,124, 50, 17,246, 50, 51,193, 32, 37,227,252, 50, 17,195,
	255,255,221, 64,222, 64,230,  0,196,128, 55,246, 50, 50,196,128,
	 50, 17, 38,240,127, 50, 51,114, 16, 17,114,  4, 17,198, 66,189,
	 35,236,126, 38,184,204,  0,  1,189, 35, 56,113,239, 17,113,251,
	 17,204,  0, 72,221, 64,198,  1,222, 64,231,  0,204,  0, 74,221,
	 64,198,128,222, 64,231,  0,124,  0, 65, 38,  3,124,  0, 64,222,
	 64,111,  0,204,  0, 50,189, 34,159,114, 16, 17,114,  4, 17,204,
	  0,  0,189, 35, 56, 57, 60, 60,204, 31,255,253, 50, 56,125, 48,
	  0,204,175,240, 48,237,  2,246,175,240,193, 16, 38,119,206,175,
	241,107,240,  0, 39,111,246,175,241,196,240,193, 96, 36,102,127,
	  0, 66, 32, 15, 48,238,  2, 58,166,  0,206, 50, 19, 58,167,  0,
	124,  0, 66,214, 66,193,  6, 37,235,204,  0,  1,189, 35, 56,113,
	239, 17,113,251, 17, 48,236,  2,221, 64,222, 64,111,  0,127,  0,
	 66,204,191,240,221, 64, 32, 21,206, 50, 19, 58,230,  0,222, 64,
	231,  0,124,  0, 66,124,  0, 65, 38,  3,124,  0, 64,214, 66,193,
	  6, 37,229,204,  0, 50,189, 34,159,114, 16, 17,114,  4, 17,204,
	  0,  0,189, 35, 56,189, 39,181,198, 82,189, 35,236, 48, 24,189,
	 36, 38, 93, 39,248, 48,230,  0,193, 65, 39, 67,193, 69, 39, 73,
	193, 71, 39, 53,193, 80, 39, 79,193, 84, 39, 85,193, 86, 39, 12,
	193, 88, 39, 81,193, 90, 39, 91,193,118, 38,209,134,  0, 32, 17,
	206, 48,  0, 22, 58,230,  0, 48,167,  1,189, 35,236, 48,166,  1,
	 76,129, 10, 37,235,198, 81, 32,177, 15,189, 64,  0, 32,174,198,
	 97,189, 35,236,204, 64,  0, 32, 36,198,101,189, 35,236,206,  4,
	 16, 60,204,187,224, 32, 46,198,112,189, 35,236,204,187,224, 32,
	 12,198,116, 32,133,198,120,189, 35,236,204,191,240,189, 38,143,
	126, 40,141,204,191,240,253, 50, 54,198,122,189, 35,236,206,  0,
	  6, 60,252, 50, 54,189, 38, 57, 56,126, 40,141,189, 34, 50,189,
	 34, 99,189, 36,153,189, 34,154, 14,127,  0, 66, 32,  9,206, 50,
	 58, 58,231,  0,124,  0, 66,214, 66,193, 60, 37,241,189, 39,246,
	 57, 60, 60, 60, 48,237,  2,150, 65, 61,231,  1,167,  0,150, 64,
	230,  3, 61,235,  0,231,  0,150, 65,230,  2, 61,235,  0,231,  0,
	236,  0, 56, 56, 56, 57,236,  0,141,215,237,  0, 57,220, 64, 38,
	  5, 79, 95,237,  2, 57,236,  4, 39,249,204,  0,  0,167,  0,237,
	  2, 32, 12,236,  4,147, 64, 35, 12,120,  0, 65,121,  0, 64,108,
	  0,150, 64, 42,238,236,  4,147, 64, 37,  5,237,  4, 13, 32,  1,
	 12,105,  3,105,  2,116,  0, 64,118,  0, 65,106,  0, 38,230, 57,
	 60, 60, 60, 60, 48,237,  4,141,180, 56, 50, 51, 56, 56, 57, 60,
	 60, 60, 60, 48,237,  4,141,165, 56, 56, 50, 51, 56, 57,150, 64,
	 42,  7,204,  0,  0,147, 64,221, 64,166,  4, 42,  7,204,  0,  0,
	163,  4,237,  4, 57, 60, 60, 60, 60, 48,237,  4,152, 64,167,  1,
	141,220,189, 41,125,109,  1, 42,  7,204,  0,  0,163,  2,237,  2,
	 56, 50, 51, 56, 56, 57, 60, 60, 60, 60, 48,237,  4,167,  1,141,
	189,189, 41,125,166,  1, 42,  7,204,  0,  0,163,  4,237,  4, 56,
	 56, 50, 51, 56, 57,236,  0,141,135,237,  0, 57,236,  0,141,181,
	237,  0, 57,236,  0,141,136,237,  0, 57,236,  0,141,200,237,  0,
	 57,  0,  0,  0,  0, 75, 71, 49, 48, 57, 86, 77, 50, 46, 49,  0,
};

static char check_size[(sizeof(boot_bin) == SIZE_BOOT) ? 1 : -1];

/* end of h45loader.c */

